/*  File: tst.c								*/

/*
 *  Copyright (C) 2005, Tres Melton          
 *
 *  Permission is hereby granted to Michael Jennings to license this code as
 *  he sees fit.  I'd prefer the GPL but he will choose the BSD. The debate
 *  is moot as this is to become a part of the Eterm project, for which he is 
 *  the primary author.  For users of this code I ask that any modifications
 *  be released back into the community but with Michael Jennings chooses the
 *  BSD license then that request has no backing in law.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL   
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER  
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN   
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * The 24 bit shading routine no longer needs the integers r, g, or b. in 
 *	pixmap.c and a patch should be submitted.
 *
 *
 *	INSTRUCTIONS FOR USE:  --->
 *
 *
 * Test parameters (adjust these in function main at the bottom of this file)
 *	width		The width of the pseudo pixmap
 *	height		The height of the pseudo pixmap
 *	red_mod		-1 = generate randomly
 *	green_mod	-1 = generate randomly
 *	blue_mod	-1 = generate randomly
 *	randomness	A seed for the random number generator (-1 = seeded by time)
 *
 * Corner cases to explore:
 *	The asm routines have two loops, one for blocks, the other for individual pixels.
 *		setting the width to all different pixels around the block should work.
 *	15/16bpp MMX handle four pixels in first loop
 *	32bpp MMX are single pixels
 *	15/16bpp SSE2 handle eight pixels in first loop
 *	32bpp SSE2 handle 2 pixels in first loop
 *
 * COLOR_MOD_MASK is used to keep the random numbers within bounds.
 *
 */

#include <time.h>
#include <stdio.h>

//  __USE_XOPEN2K seems to be the key to unlock the function posix_memalign() contrary to the man pages.
#define _XOPEN_SOURCE 600
#define _GNU_SOURCE
#define __USE_XOPEN2K
#include <stdlib.h>

#include <malloc.h>
#include <errno.h>
#include <X11/Xlib.h>

#include "cmod.h"

#include "time-it.h"
#include "align/memory.h"

#define DATA16 unsigned short
//#define COLOR_MOD_MASK 0xff
//#define COLOR_MOD_MASK 0x2ff	//  This isn't zeros then ones
#define COLOR_MOD_MASK 0x3ff	//  This is
//#define COLOR_MOD_MASK 0x30
//#define COLOR_MOD_MASK 0x4fe
#define RANDOM -1

#define TEST_VERSION_MAJOR	1
#define TEST_VERSION_MINOR	1

unsigned long long min_cycles_15_c    = -1;
unsigned long long min_cycles_16_c    = -1;
unsigned long long min_cycles_32_c    = -1;
unsigned long long min_cycles_15_simd = -1;
unsigned long long min_cycles_16_simd = -1;
unsigned long long min_cycles_32_simd = -1;

unsigned long long tot_cycles_15_c    = 0;
unsigned long long tot_cycles_16_c    = 0;
unsigned long long tot_cycles_32_c    = 0;
unsigned long long tot_cycles_15_simd = 0;
unsigned long long tot_cycles_16_simd = 0;
unsigned long long tot_cycles_32_simd = 0;

unsigned long long min_cycles_15_a    = -1;
unsigned long long min_cycles_15_u    = -1;
unsigned long long min_cycles_16_a    = -1;
unsigned long long min_cycles_16_u    = -1;

unsigned long long tot_cycles_15_a    = 0;
unsigned long long tot_cycles_15_u    = 0;
unsigned long long tot_cycles_16_a    = 0;
unsigned long long tot_cycles_16_u    = 0;

unsigned long long min_cycles_mine    = -1;     
unsigned long long min_cycles_posix   = -1;

unsigned long long tot_cycles_mine    = 0;
unsigned long long tot_cycles_posix   = 0;

unsigned get_bytes_per_line( const unsigned width, const unsigned bpp, const size_t alignment )
{
  switch( bpp )
  {
    case 15:
    case 16:
      return((( 2 * width ) + ( alignment - 1 )) & (~( alignment - 1 )));
    case 32:
      return((( 4 * width ) + ( alignment - 1 )) & (~( alignment - 1 )));
    default:
      return -1;
  }
}

int shade_tester_15bpp( int width, int height, int red_mod, int green_mod, int blue_mod, int randomness )
{
  unsigned short     *data_ptr1, *data_ptr2;
  int                 index1, index2;
  int                 error = 0;
  unsigned long long  c_cycles, simd_cycles;
  unsigned            bpl = get_bytes_per_line( width, 15, ETERM_ALIGNMENT );

  data_ptr1 = ( unsigned short* ) memalign(16, bpl * height * sizeof( unsigned short ));
  data_ptr2 = ( unsigned short* ) memalign(16, bpl * height * sizeof( unsigned short ));
  if(!( data_ptr1 && data_ptr2 ))
  {
    printf( "Failed memory allocation.\n" );
    exit( 1 );
  }
  printf( "  Testing 15 bpp mode.  " );
  if( randomness == RANDOM )
    srand( time( NULL ));
  else
    srand( randomness );
  if( red_mod == RANDOM )
    red_mod = ( COLOR_MOD_MASK & rand());
  if( green_mod == RANDOM )
    green_mod = ( COLOR_MOD_MASK & rand());
  if( blue_mod == RANDOM )
    blue_mod = ( COLOR_MOD_MASK & rand());
  for( index2 = 0; index2 < height; index2++ )
    for( index1 = 0; index1 < width; index1++ )
      data_ptr1[ index1 + bpl * index2 ] = data_ptr2[ index1 + bpl * index2 ] = 
	      0x007fff & rand();
  
  start_counter();
  shade_ximage_15_SIMD( data_ptr1, bpl, width, height, red_mod, green_mod, blue_mod );
  simd_cycles = get_counter();
  start_counter();
  shade_ximage_15_C( data_ptr2, bpl, width, height, red_mod, green_mod, blue_mod );
  c_cycles = get_counter();
  if( simd_cycles < min_cycles_15_simd )
    min_cycles_15_simd = simd_cycles;
  tot_cycles_15_simd += simd_cycles;
  if( c_cycles < min_cycles_15_c )
    min_cycles_15_c = c_cycles;
  tot_cycles_15_c += c_cycles;
  for( index2 = 0; index2 < height; index2++ )
    for( index1 = 0; index1 < width; index1++ )
      if( data_ptr1[ index1 + bpl * index2 ] != data_ptr2[ index1 + bpl * index2 ] )
      {
        if( !error )
        {
          printf( "\nData Check:            (w)           (h)             (Assembly)            (C-Code)\n" );
        }
        error++;
        printf( "    Mismatch:  index1:  %-3d  index2:  %-3d  data_ptr1:  0x%04x   data_ptr2:  0x%04x\n", 
                     index1, index2, data_ptr1[ index1 + bpl * index2 ], data_ptr2[ index1 + bpl * index2 ] );
        if( error > 10 )
          exit( 1 );
      }
  if( !error )
    printf( "Data compared Identically.                                 (%dx%d=%d pixels)\n", width, height, width * height );
  else
    printf( "Data Compare had %d mismatch(s).  (%dx%d=%d pixels)\n", error, width, height, width * height );
  printf( "    C Cycles:    %15lld   %5lld cycles/pixel\n    SIMD Cycles: %15lld   %5lld cycles/pixel  (%3d%% reduction)\n", 
               c_cycles, ( c_cycles / ( width * height )), simd_cycles, ( simd_cycles / ( width * height )),
               ((int)( 100 - (( simd_cycles * 100 ) / c_cycles ))));
  free( data_ptr1 );
  free( data_ptr2 );
  return error;
}

int shade_tester_16bpp( int width, int height, int red_mod, int green_mod, int blue_mod, int randomness )
{
  unsigned short     *data_ptr1, *data_ptr2;
  int                 index1, index2;
  int                 error = 0;
  unsigned long long  c_cycles, simd_cycles;
  unsigned            bpl = get_bytes_per_line( width, 16, ETERM_ALIGNMENT );

  data_ptr1 = ( unsigned short* ) malloc_aligned( bpl * height * sizeof( unsigned short ), 16 );
  data_ptr2 = ( unsigned short* ) malloc( bpl * height * sizeof( unsigned short ));
  if(!( data_ptr1 && data_ptr2 ))
  {
    printf( "Failed memory allocation.\n" );
    exit( 1 );
  }
  printf( "  Testing 16 bpp mode.  " );
  if( randomness == RANDOM )
    srand( time( NULL ));
  else
    srand( randomness );
  if( red_mod == RANDOM )
    red_mod = ( COLOR_MOD_MASK & rand());
  if( green_mod == RANDOM )
    green_mod = ( COLOR_MOD_MASK & rand());
  if( blue_mod == RANDOM )
    blue_mod = ( COLOR_MOD_MASK & rand());
  for( index2 = 0; index2 < height; index2++ )
    for( index1 = 0; index1 < width; index1++ )
      data_ptr1[ index1 + bpl * index2 ] = data_ptr2[ index1 + bpl * index2 ] = 0x00ffff & rand();

  start_counter();
  shade_ximage_16_SIMD( data_ptr1, bpl, width, height, red_mod, green_mod, blue_mod );
  simd_cycles = get_counter();
  start_counter();
  shade_ximage_16_C( data_ptr2, bpl, width, height, red_mod, green_mod, blue_mod );
  c_cycles = get_counter();
  if( simd_cycles < min_cycles_16_simd )
    min_cycles_16_simd = simd_cycles;
  tot_cycles_16_simd += simd_cycles;
  if( c_cycles < min_cycles_16_c )
    min_cycles_16_c = c_cycles;
  tot_cycles_16_c += c_cycles;
  for( index2 = 0; index2 < height; index2++ )
    for( index1 = 0; index1 < width; index1++ )
      if( data_ptr1[ index1 + bpl * index2 ] != data_ptr2[ index1 + bpl * index2 ] )
      {
        if( !error )
        {
          printf( "\nData Check:            (w)           (h)             (Assembly)            (C-Code)\n" );
        }
        error++;
        printf( "    Mismatch:  index1:  %-3d  index2:  %-3d  data_ptr1:  0x%04x   data_ptr2:  0x%04x\n", 
                     index1, index2, data_ptr1[ index1 + width * index2 ], data_ptr2[ index1 + width * index2 ] );
        if( error > 10 )
          exit( 1 );
      }
  if( !error )
    printf( "Data compared Identically.                                 (%dx%d=%d pixels)\n", width, height, width * height );
  else
    printf( "Data compare had %d mismatch(s).  (%dx%d=%d pixels)\n", error, width, height, width * height );
  printf( "    C Cycles:    %15lld   %5lld cycles/pixel\n    SIMD Cycles: %15lld   %5lld cycles/pixel  (%3d%% reduction)\n", 
               c_cycles, ( c_cycles / ( width * height )), simd_cycles, ( simd_cycles / ( width * height )),
               ((int)( 100 - (( simd_cycles * 100 ) / c_cycles ))));
  free_aligned( data_ptr1 );
  free( data_ptr2 );
  return error;
}

int shade_tester_32bpp( int width, int height, int red_mod, int green_mod, int blue_mod, int randomness )
{
  unsigned int       *data_ptr1, *data_ptr2;
  int                 index1, index2;
  int                 error = 0;
  int                 round_error = 0;
  unsigned long long  c_cycles, simd_cycles;
  unsigned            bpl = get_bytes_per_line( width, 32, ETERM_ALIGNMENT );

  data_ptr1 = ( unsigned int* ) memalign(16, bpl * height * sizeof( unsigned int ));
  data_ptr2 = ( unsigned int* ) memalign(16, bpl * height * sizeof( unsigned int ));
  if(!( data_ptr1 && data_ptr2 ))
  {
    printf( "Failed memory allocation.\n" );
    exit( 1 );
  }
  printf( "  Testing 32 bpp mode.  " );
  if( randomness == RANDOM )
    srand( time( NULL ));
  else
    srand( randomness );
  if( red_mod == RANDOM )
    red_mod = ( COLOR_MOD_MASK & rand());
  if( green_mod == RANDOM )
    green_mod = ( COLOR_MOD_MASK & rand());
  if( blue_mod == RANDOM )
    blue_mod = ( COLOR_MOD_MASK & rand());
  for( index2 = 0; index2 < height; index2++ )
    for( index1 = 0; index1 < width; index1++ )
      data_ptr1[ index1 + bpl * index2 ] = data_ptr2[ index1 + bpl * index2 ] = 
	      0x00ffffff & rand();

  start_counter();
  shade_ximage_32_SIMD( data_ptr1, bpl, width, height, red_mod, green_mod, blue_mod );
  simd_cycles = get_counter();
  start_counter();
  shade_ximage_32_C( data_ptr2, bpl, width, height, red_mod, green_mod, blue_mod );
  c_cycles = get_counter();
  if( simd_cycles < min_cycles_32_simd )
    min_cycles_32_simd = simd_cycles;
  tot_cycles_32_simd += simd_cycles;
  if( c_cycles < min_cycles_32_c )
    min_cycles_32_c = c_cycles;
  tot_cycles_32_c += c_cycles;
  for( index2 = 0; index2 < height; index2++ )
    for( index1 = 0; index1 < width; index1++ )
      if( data_ptr1[ index1 + bpl * index2 ] != data_ptr2[ index1 + bpl * index2 ] )
      {
        register int temp1 = data_ptr1[ index1 + bpl * index2 ];
        register int temp2 = data_ptr2[ index1 + bpl * index2 ];

        if((((  temp1 & 0x000000ff )         - (  temp2 & 0x000000ff )         >= -1 )    && 
           (((  temp1 & 0x000000ff )         - (  temp2 & 0x000000ff )         <=  1 )))  && 
           (((( temp1 & 0x0000ff00 ) >>  8 ) - (( temp2 & 0x0000ff00 ) >>  8 ) >= -1 )    && 
           (((( temp1 & 0x0000ff00 ) >>  8 ) - (( temp2 & 0x0000ff00 ) >>  8 ) <=  1 )))  && 
           (((( temp1 & 0x00ff0000 ) >> 16 ) - (( temp2 & 0x00ff0000 ) >> 16 ) >= -1 )    && 
           (((( temp1 & 0x00ff0000 ) >> 16 ) - (( temp2 & 0x00ff0000 ) >> 16 ) <=  1 ))))
        {
          round_error++;
        }
        else
        {
          if( !error )
          {
            printf( "\nData Check:            (w)           (h)               (Assembly)                 (C-Code)\n" );
          }
          error++;
          printf( "    Mismatch:  index1:  %-3d  index2:  %-3d  data_ptr1:  0x%08x   data_ptr2:  0x%08x\n", 
                       index1, index2, data_ptr1[ index1 + width * index2 ], data_ptr2[ index1 + width * index2 ] );

          printf(  "        r1 %d - r2 %d        g1 %d - g2 %d        b1 %d - b2 %d\n",
          	(( temp1 & 0x00ff0000 ) >> 16 ), (( temp2 & 0x00ff0000 ) >> 16 ),
                (( temp1 & 0x0000ff00 ) >>  8 ), (( temp2 & 0x0000ff00 ) >>  8 ),
                (( temp1 & 0x000000ff ) >>  0 ), (( temp2 & 0x000000ff ) >>  0 ));
          if( error > 10 )
            exit( 1 );
        }
      }
  if( !error )
    if( round_error )
      printf( "Data compared within SIMD rounding error %8d time(s). (%dx%d=%d pixels)\n", round_error, width, height, width * height );
    else
      printf( "Data compared Identically.                                 (%dx%d=%d pixels)\n", width, height, width * height );
  else
    printf( "Data compare had %d mismatch(s) and %d rounding error(s).  (%dx%d pixels)\n", error, round_error, width, height );
  printf( "    C Cycles:    %15lld   %5lld cycles/pixel\n    SIMD Cycles: %15lld   %5lld cycles/pixel  (%3d%% reduction)\n", 
               c_cycles, ( c_cycles / ( width * height )), simd_cycles, ( simd_cycles / ( width * height )),
               ((int)( 100 - (( simd_cycles * 100 ) / c_cycles ))));
  free( data_ptr1 );
  free( data_ptr2 );
  return error;
}

#if SIMD_USE_ALIGNED
int aligned_v_unaligned_tester15( int width, int height, int red_mod, int green_mod, int blue_mod, int randomness )
{
  unsigned short     *data_ptr1, *data_ptr2;
  int                 index1, index2;
  int                 error = 0;
  unsigned long long  u_cycles, a_cycles;
  unsigned            bpl = get_bytes_per_line( width, 15, ETERM_ALIGNMENT );

  data_ptr1 = ( unsigned short* ) malloc_aligned( bpl * height * sizeof( unsigned short ), 16 );
  if( !data_ptr1 )
  {
    printf( "Failed Aligned allocation.\n" );
    exit( 1 );
  }
  data_ptr2 = ( unsigned short* ) malloc( bpl * height * sizeof( unsigned short ));
  if( !data_ptr2 )
  {
    printf( "Failed memory allocation.\n" );
    exit( 1 );
  }
  printf( "  Testing 15 bpp unaligned v aligned mode.  " );
  if( randomness == RANDOM )
    srand( time( NULL ));
  else
    srand( randomness );
  if( red_mod == RANDOM )
    red_mod = ( COLOR_MOD_MASK & rand());
  if( green_mod == RANDOM )
    green_mod = ( COLOR_MOD_MASK & rand());
  if( blue_mod == RANDOM )
    blue_mod = ( COLOR_MOD_MASK & rand());
  for( index2 = 0; index2 < height; index2++ )
    for( index1 = 0; index1 < width; index1++ )
      data_ptr1[ index1 + bpl * index2 ] = data_ptr2[ index1 + bpl * index2 ] = 0x00ffff & rand();

#if   ETERM_SIMD == ETERM_SIMD_SSE2  
#  if ETERM_ARCH == ETERM_ARCH_x86
  start_counter();
  shade_ximage_15_x86_sse2_A( data_ptr1, bpl, width, height, red_mod, green_mod, blue_mod );
  a_cycles = get_counter();
  start_counter();
  shade_ximage_15_x86_sse2_U( data_ptr2, bpl, width, height, red_mod, green_mod, blue_mod );
  u_cycles = get_counter();
#elif ETERM_ARCH == ETERM_ARCH_x86_64
  start_counter();
  shade_ximage_15_x86_64_sse2_A( data_ptr1, bpl, width, height, red_mod, green_mod, blue_mod );
  a_cycles = get_counter();
  start_counter();
  shade_ximage_15_x86_64_sse2_U( data_ptr2, bpl, width, height, red_mod, green_mod, blue_mod );
  u_cycles = get_counter();
#  endif
#endif
  if( u_cycles < min_cycles_15_u )
    min_cycles_15_u = u_cycles;
  tot_cycles_15_u += u_cycles;
  if( a_cycles < min_cycles_15_a )
    min_cycles_15_a = a_cycles;
  tot_cycles_15_a += a_cycles;
  for( index2 = 0; index2 < height; index2++ )
    for( index1 = 0; index1 < width; index1++ )
      if( data_ptr1[ index1 + bpl * index2 ] != data_ptr2[ index1 + bpl * index2 ] )
      {
        if( !error )
        {
          printf( "\nData Check:            (w)           (h)             (Assembly)            (C-Code)\n" );
        }
        error++;
        printf( "    Mismatch:  index1:  %-3d  index2:  %-3d  data_ptr1:  0x%04x   data_ptr2:  0x%04x\n", 
                     index1, index2, data_ptr1[ index1 + width * index2 ], data_ptr2[ index1 + width * index2 ] );
        if( error > 10 )
          exit( 1 );
      }
  if( !error )
    printf( "Data compared Identically.             (%dx%d=%d pixels)\n", width, height, width * height );
  else
    printf( "Data compare had %d mismatch(s).     (%dx%d=%d pixels)\n", error, width, height, width * height );
  printf( "    Unaligned:   %15lld   %5lld cycles/pixel\n    Aligned:     %15lld   %5lld cycles/pixel  (%3d%% reduction)\n", 
               u_cycles, ( u_cycles / ( width * height )), a_cycles, ( a_cycles / ( width * height )),
               ((int)(( 100 - (( a_cycles * 100 ) / u_cycles )))));
  free_aligned( data_ptr1 );
  free( data_ptr2 );
  return error;
}

int aligned_v_unaligned_tester16( int width, int height, int red_mod, int green_mod, int blue_mod, int randomness )
{
  unsigned short     *data_ptr1, *data_ptr2;
  int                 index1, index2;
  int                 error = 0;
  unsigned long long  u_cycles, a_cycles;
  unsigned            bpl = get_bytes_per_line( width, 16, ETERM_ALIGNMENT );

  data_ptr1 = ( unsigned short* ) malloc_aligned( bpl * height * sizeof( unsigned short ), 16 );
  if( !data_ptr1 )
  {
    printf( "Failed Aligned allocation.\n" );
    exit( 1 );
  }
  data_ptr2 = ( unsigned short* ) malloc( bpl * height * sizeof( unsigned short ));
  if( !data_ptr2 )
  {
    printf( "Failed memory allocation.\n" );
    exit( 1 );
  }
  printf( "  Testing 16 bpp unaligned v aligned mode.  " );
  if( randomness == RANDOM )
    srand( time( NULL ));
  else
    srand( randomness );
  if( red_mod == RANDOM )
    red_mod = ( COLOR_MOD_MASK & rand());
  if( green_mod == RANDOM )
    green_mod = ( COLOR_MOD_MASK & rand());
  if( blue_mod == RANDOM )
    blue_mod = ( COLOR_MOD_MASK & rand());
  for( index2 = 0; index2 < height; index2++ )
    for( index1 = 0; index1 < width; index1++ )
      data_ptr1[ index1 + bpl * index2 ] = data_ptr2[ index1 + bpl * index2 ] = 0x00ffff & rand();

#if     ETERM_SIMD == ETERM_SIMD_SSE2
#  if   ETERM_ARCH == ETERM_ARCH_x86
  start_counter();
  shade_ximage_16_x86_sse2_A( data_ptr1, bpl, width, height, red_mod, green_mod, blue_mod );
  a_cycles = get_counter();
  start_counter();
  shade_ximage_16_x86_sse2_U( data_ptr2, bpl, width, height, red_mod, green_mod, blue_mod );
  u_cycles = get_counter();
#  elif ETERM_ARCH == ETERM_ARCH_x86_64
  start_counter();
  shade_ximage_16_x86_64_sse2_A( data_ptr1, bpl, width, height, red_mod, green_mod, blue_mod );
  a_cycles = get_counter();
  start_counter();
  shade_ximage_16_x86_64_sse2_U( data_ptr2, bpl, width, height, red_mod, green_mod, blue_mod );
  u_cycles = get_counter();
#  endif
#endif
  if( u_cycles < min_cycles_16_u )
    min_cycles_16_u = u_cycles;
  tot_cycles_16_u += u_cycles;
  if( a_cycles < min_cycles_16_a )
    min_cycles_16_a = a_cycles;
  tot_cycles_16_a += a_cycles;
  for( index2 = 0; index2 < height; index2++ )
    for( index1 = 0; index1 < width; index1++ )
      if( data_ptr1[ index1 + bpl * index2 ] != data_ptr2[ index1 + bpl * index2 ] )
      {
        if( !error )
        {
          printf( "\nData Check:            (w)           (h)             (Assembly)            (C-Code)\n" );
        }
        error++;
        printf( "    Mismatch:  index1:  %-3d  index2:  %-3d  data_ptr1:  0x%04x   data_ptr2:  0x%04x\n", 
                     index1, index2, data_ptr1[ index1 + width * index2 ], data_ptr2[ index1 + width * index2 ] );
        if( error > 10 )
          exit( 1 );
      }
  if( !error )
    printf( "Data compared Identically.             (%dx%d=%d pixels)\n", width, height, width * height );
  else
    printf( "Data compare had %d mismatch(s).     (%dx%d=%d pixels)\n", error, width, height, width * height );
  printf( "    Unaligned:   %15lld   %5lld cycles/pixel\n    Aligned:     %15lld   %5lld cycles/pixel  (%3d%% reduction)\n", 
               u_cycles, ( u_cycles / ( width * height )), a_cycles, ( a_cycles / ( width * height )),
               ((int)( 100 - (( a_cycles * 100 ) / u_cycles ))));
  free_aligned( data_ptr1 );
  free( data_ptr2 );
  return error;
}

int my_memory_v_posix( int width, int height, int red_mod, int green_mod, int blue_mod, int randomness )
{
  unsigned short     *data_ptr1, *data_ptr2;
  int                 index1, index2;
  int                 error = 0;
  unsigned long long  p_cycles, m_cycles;
  unsigned            bpl = get_bytes_per_line( width, 15, ETERM_ALIGNMENT );
//  XImage	 *image_ptr;

  printf( "  Testing posix v me memory allocation mode.  " );

  start_counter();
  data_ptr1 = ( unsigned short* ) malloc_aligned( bpl * height * sizeof( unsigned short ), 16 );
  m_cycles = get_counter();
  if( !data_ptr1 )
  {
    printf( "Failed Aligned allocation.\n" );
    exit( 1 );
  }
  start_counter();
  error = posix_memalign((void**) &data_ptr2, 16, ( bpl * height * sizeof( unsigned short )));
  p_cycles = get_counter();
  if( error == EINVAL )
  {
    printf( "Invalid alignment requested.\n" );
    exit( 1 );
  }
  else if( error == ENOMEM )
  {
    printf( "Not enough memory.\n" );
    exit( 1 );
  }

  if( randomness == RANDOM )
    srand( time( NULL ));
  else
    srand( randomness );
  if( red_mod == RANDOM )
    red_mod = ( COLOR_MOD_MASK & rand());
  if( green_mod == RANDOM )
    green_mod = ( COLOR_MOD_MASK & rand());
  if( blue_mod == RANDOM )
    blue_mod = ( COLOR_MOD_MASK & rand());
  for( index2 = 0; index2 < height; index2++ )
    for( index1 = 0; index1 < width; index1++ )
      data_ptr1[ index1 + bpl * index2 ] = data_ptr2[ index1 + bpl * index2 ] = 0x00ffff & rand();

#if     ETERM_SIMD == ETERM_SIMD_SSE2
#  if   ETERM_ARCH == ETERM_ARCH_x86
  shade_ximage_15_x86_sse2_A( data_ptr1, bpl, width, height, red_mod, green_mod, blue_mod );
  shade_ximage_16_x86_sse2_A( data_ptr2, bpl, width, height, red_mod, green_mod, blue_mod );
#elif   ETERM_ARCH == ETERM_ARCH_x86_64
  shade_ximage_15_x86_64_sse2_A( data_ptr1, bpl, width, height, red_mod, green_mod, blue_mod );
  shade_ximage_15_x86_64_sse2_A( data_ptr2, bpl, width, height, red_mod, green_mod, blue_mod );
#  endif
#endif

  if( m_cycles < min_cycles_mine )
    min_cycles_mine = m_cycles;
  tot_cycles_mine += m_cycles;
  if( p_cycles < min_cycles_posix )
    min_cycles_posix = p_cycles;
  tot_cycles_posix += p_cycles;
  for( index2 = 0; index2 < height; index2++ )
    for( index1 = 0; index1 < width; index1++ )
      if( data_ptr1[ index1 + bpl * index2 ] != data_ptr2[ index1 + bpl * index2 ] )
      {
        if( !error )
        {
          printf( "\nData Check:            (w)           (h)             (Assembly)            (C-Code)\n" );
        }
        error++;
        printf( "    Mismatch:  index1:  %-3d  index2:  %-3d  data_ptr1:  0x%04x   data_ptr2:  0x%04x\n", 
                     index1, index2, data_ptr1[ index1 + bpl * index2 ], data_ptr2[ index1 + bpl * index2 ] );
        if( error > 10 )
          exit( 1 );
      }
  if( !error )
    printf( "Data compared Identically.           (%dx%d=%d pixels)\n", width, height, width * height );
  else
    printf( "Data compare had %d mismatch(s).     (%dx%d=%d pixels)\n", error, width, height, width * height );
  printf( "    Posix:       %15lld   \n    Mine:        %15lld                       (%3d%% reduction)\n", 
               p_cycles, m_cycles, 
               ((int)( 100 - (( p_cycles * 100 ) / m_cycles ))));
  free_aligned( data_ptr1 );
  free( data_ptr2 );
  return error;
}
#endif

int info_dump( void )
{
  printf( "-----------------------------------------------------------\n" );
  printf( "CPU info:\n" );
  system( "cat /proc/cpuinfo" );
  printf( "-----------------------------------------------------------\n" );
  printf( "Load info:\n" );
  system( "cat /proc/loadavg" );
  printf( "-----------------------------------------------------------\n" );
  printf( "System info:\n" );
  system( "uname -a" );
//  printf( "-----------------------------------------------------------\n" );
//  system( "cat " );
//  system( "" );
  return 1;
}

int macro_dump( void )
{
#if     ETERM_ARCH == ETERM_ARCH_x86
#  if   ETERM_SIMD == ETERM_SIMD_MMX
  printf( "x86     MMX Routines.\n" );
#  elif ETERM_SIMD == ETERM_SIMD_SSE
  printf( "x86     SSE Routines.\n" );
#  elif ETERM_SIMD == ETERM_SIMD_SSE2
  printf( "x86    SSE2 Routines.\n" );
#  else
  printf( "          C Routines.\n" );
#  endif
#elif   ETERM_ARCH == ETERM_ARCH_x86_64
#  if   ETERM_SIMD == ETERM_SIMD_SSE2
  printf( "x86-64 SSE2 Routines.\n" );
#  else		/*  Other, lesser, combinations make no sense  */
  printf( "          C Routines.\n" );
#  endif
#else
  printf( "          C Routines.\n" );
#endif
  return 1;
}

int main( void )
{
  int error;
  int index;
  int count  = 10;
//  int count  = 5000;
//  int width = 7;
//  int height = 3;
//  int width  = 5011;
//  int height = 5099;
  int width  = 4800;
  int height = 4800;
  int pixels = width * height;
  int red_mod   = RANDOM;
  int green_mod = RANDOM;
  int blue_mod  = RANDOM;
  int r_seed = RANDOM;
//  int red_mod   = 0x050;
//  int green_mod = 0x060;
//  int blue_mod  = 0x070;
//  int r_seed = 123;

  access_counter_init();
  printf( "===========================================================\n" );
  printf( "Starting SIMD testing/profiling.\n" );
  printf( "  Version:  %d.%d\n", TEST_VERSION_MAJOR, TEST_VERSION_MINOR );
  macro_dump();
  printf( "Start time:\n" );
  system( "date" );
  info_dump();
  printf( "-----------------------------------------------------------\n" );

#if 1
  index = 1;
  while( index <= count )
  {
    printf( "Iteration %d\n", index );
#if 1
    if(( error = shade_tester_15bpp( width, height, red_mod, green_mod, blue_mod, r_seed )))
    {
      printf( "%d error(s) occured in iteration %d while shading a 15bpp pixmap.\n", error, index );
      break;
    }
#endif
#if 0
    if(( error = shade_tester_16bpp( width, height, red_mod, green_mod, blue_mod, r_seed )))
    {
      printf( "%d error(s) occured in iteration %d while shading a 16bpp pixmap.\n", error, index );
      break;
    }
#endif
#if 1 
    if(( error = shade_tester_32bpp( width, height, red_mod, green_mod, blue_mod, r_seed )))
    {
      printf( "%d error(s) occured in iteration %d while shading a 32bpp pixmap.\n", error, index );
      break;
    }
#endif
    index++;
  }
//exit(0);
  if( error )
    return( error );
  printf( "\nMinimum cycles per image (per pixel):\n" );
  printf( "bpp \\ routine          C (/pix)          SIMD (/pix) \n" );
  printf( "15          %12lld (%4lld)  %12lld (%4lld)   (%3d%% reduction)\n", min_cycles_15_c, min_cycles_15_c / pixels, min_cycles_15_simd, 
           min_cycles_15_simd / pixels, ((int) ( 100 - (( min_cycles_32_simd * 100 ) / min_cycles_32_c ))));
  printf( "16          %12lld (%4lld)  %12lld (%4lld)   (%3d%% reduction)\n", min_cycles_16_c, min_cycles_16_c / pixels, min_cycles_16_simd, 
           min_cycles_16_simd / pixels, ((int) ( 100 - (( min_cycles_32_simd * 100 ) / min_cycles_32_c ))));
  printf( "32          %12lld (%4lld)  %12lld (%4lld)   (%3d%% reduction)\n", min_cycles_32_c, min_cycles_32_c / pixels, min_cycles_32_simd, 
           min_cycles_32_simd / pixels, ((int) ( 100 - (( min_cycles_32_simd * 100 ) / min_cycles_32_c ))));
  printf( "\nAverage cycles per image (per pixel):\n" );
  printf( "bpp \\ routine          C (/pix)          SIMD (/pix) \n" );
  printf( "15          %12lld (%4lld)  %12lld (%4lld)   (%3d%% reduction)\n", ( tot_cycles_15_c / count ), (( tot_cycles_15_c / pixels ) / count ), 
        ( tot_cycles_15_simd / count ), (( tot_cycles_15_simd / pixels ) / count ), ((int) ( 100 - (( tot_cycles_15_simd * 100 ) / tot_cycles_15_c ))));
  printf( "16          %12lld (%4lld)  %12lld (%4lld)   (%3d%% reduction)\n", ( tot_cycles_16_c / count ), (( tot_cycles_16_c / pixels ) / count ), 
        ( tot_cycles_16_simd / count ), (( tot_cycles_16_simd / pixels ) / count ), ((int) ( 100 - (( tot_cycles_16_simd * 100 ) / tot_cycles_16_c ))));
  printf( "32          %12lld (%4lld)  %12lld (%4lld)   (%3d%% reduction)\n", ( tot_cycles_32_c / count ), (( tot_cycles_32_c / pixels ) / count ), 
        ( tot_cycles_32_simd / count ), (( tot_cycles_32_simd / pixels ) / count ), ((int) ( 100 - (( tot_cycles_32_simd * 100 ) / tot_cycles_32_c ))));
  printf( "-----------------------------------------------------------\n" );
#endif

  index = 1;
  while( index <= count )
  {
    //  Test aligned vs. unaligned stuff here.  Also test with the various memory handlers.
    printf( "Iteration %d\n", index );
#if     ETERM_SIMD == ETERM_SIMD_SSE2 && SIMD_USE_ALIGNED
    if(( error = aligned_v_unaligned_tester15( width, height, red_mod, green_mod, blue_mod, r_seed )))
    {
      printf( "%d error(s) occured in iteration %d while shading a 15bpp aligned pixmap.\n", error, index );
      break;
    }
    if(( error = aligned_v_unaligned_tester16( width, height, red_mod, green_mod, blue_mod, r_seed )))
    {
      printf( "%d error(s) occured in iteration %d while shading a 16bpp aligned pixmap.\n", error, index );
      break;
    }
#endif
#if SIMD_USE_ALIGNED
    if(( error = my_memory_v_posix( width, height, red_mod, green_mod, blue_mod, r_seed )))
    {
      printf( "%d error(s) occured in iteration %d while .....   shading a 15bpp aligned pixmap.\n", error, index );
      break;
    }
#endif
    index++;
  }
  if( error )
    return( error );
  printf( "-----------------------------------------------------------\n" );
#if     ETERM_SIMD == ETERM_SIMD_SSE2 && ETERM_ARCH == ETERM_ARCH_x86_64 && SIMD_USE_ALIGNED
  printf( "\nMinimum cycles per image (per pixel):\n" );
  printf( "bpp \\ routine  Unaligned (/pix)       Aligned (/pix) \n" );
  printf( "15          %12lld (%4lld)  %12lld (%4lld)   (%3d%% reduction)\n", min_cycles_15_u, min_cycles_15_u / pixels, min_cycles_15_a, 
           min_cycles_15_a / pixels, ((int) ( 100 - (( min_cycles_15_a * 100 ) / min_cycles_15_u ))));
  printf( "16          %12lld (%4lld)  %12lld (%4lld)   (%3d%% reduction)\n", min_cycles_16_u, min_cycles_16_u / pixels, min_cycles_16_a, 
           min_cycles_16_a / pixels, ((int) ( 100 - (( min_cycles_16_a * 100 ) / min_cycles_16_u ))));
  printf( "\nAverage cycles per image (per pixel):\n" );
  printf( "bpp \\ routine  Unaligned (/pix)       Aligned (/pix) \n" );
  printf( "15          %12lld (%4lld)  %12lld (%4lld)   (%3d%% reduction)\n", ( tot_cycles_15_u / count ), (( tot_cycles_15_u / pixels ) / count ), 
        ( tot_cycles_15_a / count ), (( tot_cycles_15_a / pixels ) / count ), ((int) ( 100 - (( tot_cycles_15_a * 100 ) / tot_cycles_15_u ))));
  printf( "16          %12lld (%4lld)  %12lld (%4lld)   (%3d%% reduction)\n", ( tot_cycles_16_u / count ), (( tot_cycles_16_u / pixels ) / count ), 
        ( tot_cycles_16_a / count ), (( tot_cycles_16_a / pixels ) / count ), ((int) ( 100 - (( tot_cycles_16_a * 100 ) / tot_cycles_16_u ))));
  printf( "-----------------------------------------------------------\n" );
#endif
#if SIMD_USE_ALIGNED
  printf( "\nCycles per aligned allocation:\n" );
  printf( "               My_memory                Posix\n" );
  printf( "Minimum     %12lld         %12lld          (%3d%% reduction)\n", ( min_cycles_mine / count ),
        ( min_cycles_posix / count ), ((int) ( 100 - (( min_cycles_posix * 100 ) / min_cycles_mine ))));
  printf( "Average     %12lld         %12lld          (%3d%% reduction)\n", ( tot_cycles_mine / count ),
        ( tot_cycles_posix / count ), ((int) ( 100 - (( tot_cycles_posix * 100 ) / tot_cycles_mine ))));
#endif
  printf( "\n" );
  printf( "-----------------------------------------------------------\n" );
  printf( "End time:\n" );
  system( "date" );
  printf( "===========================================================\n" );
  return error;
}

















